
// includes
#include <fstream>
#include <iostream>

#include "bleifrei/gfx/opengl.hpp"
#include "bleifrei/gfx/EffectManager.hpp"
#include "bleifrei/math/Matrix_op.hpp"
#include "bleifrei/img/TGAImage.hpp"
#include "bleifrei/input/Keyboard.hpp"
#include "bleifrei/input/Mouse.hpp"
#include "bleifrei/io/ConfigFile.hpp"
#include "bleifrei/io/Console.hpp"
#include "bleifrei/platform/System.hpp"
#include "bleifrei/platform/Window.hpp"
#include "bleifrei/tex/TextureManager.hpp"

#include "gui/GUI.hpp"

#include "global.hpp"

using namespace bleifrei::gfx;
using namespace bleifrei::img;
using namespace bleifrei::input;
using namespace bleifrei::io;
using namespace bleifrei::math;
using namespace bleifrei::platform;
using namespace bleifrei::tex;
using namespace bleifrei;
using namespace std;

// global window size
int g_width				= 1024;
int g_height			= 768;
int g_bpp				= 32;
bool g_fullscreen		= false;
bool g_postprocessing	= false;

// pointer to global gui
GUI *g_gui = 0;

// global camera variables
const float MaxCameraDistance	= 1000.0f;
const float MinCameraDistance	= 300.0f;
float g_cameraDistance			= MaxCameraDistance;
float g_cameraTargetDistance	= 500.0f;
Vector3f g_cameraPosition;
Matrix4x4f g_modelMatrix;
Matrix4x4f g_projectionMatrix;
Matrix4x4f g_viewMatrix;


// global animation parameters
float g_time					= 0.0f;
float g_rotation				= 0.0f;
float g_targetRotation			= M_PI * (5 / 4.0f);

// window event listener
class EventListener :
	public WindowEventListener,
	public KeyboardEventListener
{
public:
	void onClose(void)
	{
		g_system.shutdown();
	}
	void onResize(int width, int height)
	{
		g_width = width;
		g_height = height;
		glViewport(0, 0, width, height);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(40.0f, width/(float)height, 1, 2000.0f);
		glMatrixMode(GL_MODELVIEW);

		g_gui->resize(width, height);
	}
	void onKeyTyped(char c) {
		g_gui->keyTyped(c);
	}
};

// global event listener
EventListener g_eventListener;

//-------------------------------------------------------------------------------------------------
// screenshot
//		Save a screenshot to file. Each session starts at shot0.tga.
//-------------------------------------------------------------------------------------------------
void
screenshot(void)
{
	static int count = 0;
	char buffer[20];

	string filename = "shot";
	filename += _itoa(count++, buffer, 10);
	filename += ".tga";

	Image image(g_width, g_height, ImageFormatB8G8R8);
	glReadPixels(0, 0, g_width, g_height, GL_BGR, GL_UNSIGNED_BYTE, image.getData());

	((TGAImage *)&image)->save(ofstream(filename.c_str(), ios::binary), ImageCompressionRaw);
}

//-------------------------------------------------------------------------------------------------
// setLoadingProgressMax
//		Loading bar help function. Redirects call to GUI.
//-------------------------------------------------------------------------------------------------
void
setLoadingProgressMax(int max)
{
	g_gui->setLoadingProgressMax(max);
}

//-------------------------------------------------------------------------------------------------
// progressCallback
//		Loading bar callback function. Redraws GUI after each loading task has finished.
//-------------------------------------------------------------------------------------------------
void
progressCallback(int progress)
{
	g_system.update();

	g_gui->setLoadingProgress(progress);

	glClearColor(1,1,1,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();

	g_gui->render();

	g_window.swapBuffers();
}

//-------------------------------------------------------------------------------------------------
// setLoadingText
//		Sets name of current loading task. Loading bar help function. Redirects call to GUI.
//-------------------------------------------------------------------------------------------------
void
setLoadingText(const string &text)
{
	g_gui->setLoadingText(text);
}

System &g_system = System::getInstance();
Console &g_console = Console::getInstance();
Window &g_window = Window::getInstance();
Keyboard &g_keyboard = Keyboard::getInstance();
Mouse &g_mouse = Mouse::getInstance();
TextureManager &g_textureManager = TextureManager::getInstance();
EffectManager &g_effectManager = EffectManager::getInstance();

//-------------------------------------------------------------------------------------------------
// applicationMain
//		Main loop of application.
//-------------------------------------------------------------------------------------------------
void
applicationMain(void)
{
	bool initialized = true;

	// read configuration
	try
	{

		ConfigFile configuration(ifstream("fxstudio.ini"));

		configuration.selectSection("Screen");
		configuration.get("width", g_width);
		configuration.get("height", g_height);
		configuration.get("bpp", g_bpp);
		configuration.get("fullscreen", g_fullscreen);

		configuration.selectSection("Scene");
		configuration.get("effectfile", g_effectFilename);
		configuration.get("postprocessing", g_postprocessing);
	}
	catch (...)
	{
	}

	// initialize subsystems
	initialized |= g_system.init();
	initialized |= g_console.init();
	initialized |= g_window.init("EDA425: Laboration Project", g_width, g_height, g_bpp, g_fullscreen);
	initialized |= g_keyboard.init();
	initialized |= g_mouse.init();

	// initialize OpenGL functionality for version > 1.1
	init_opengl();

	// initialize OpenGL extensions
	init_GL_EXT_texture_filter_anisotropic();

	// initialize managers
	initialized |= g_textureManager.init("media");
	initialized |= g_effectManager.init("fx");

	if (initialized)
	{
		// initialize GUI and set state to loading state
		g_gui = new GUI();
		g_gui->setState(GUIStateLoading);
		g_gui->setLoadingProgress(0);
		g_gui->setLoadingText("Initializing application...");

		// Reset progress bar
		progressCallback(0);

		// initialize rendering (render.cpp)
		initRender();

		// mouse variables
		bool mouseRelative = false;
		int mouseLastX, mouseLastY;

		g_mouse.setMode(MouseModeAbsolute);

		glEnable(GL_DEPTH_TEST);

		// set callback handler
		g_window.setEventListener(&g_eventListener);
		g_keyboard.setEventListener(&g_eventListener);

		g_system.resetTimer();

		// initialization finished start main loop
		g_gui->setState(GUIStateMain);
		while (g_system.update())
		{
			// exit?
			if (g_keyboard.isDown(Keyboard::Escape)) break;

			// save screenshot
			static bool togglePrintScreen = false;
			if (g_keyboard.isDown(Keyboard::PrintScreen))
			{
				if (!togglePrintScreen)
				{
					screenshot();
					togglePrintScreen = true;
				}
			}
			else
			{
				togglePrintScreen = false;
			}

			// increase total elapsed time
			g_time += g_system.getFrameTime();

			// update subsystems
			g_keyboard.update();
			g_mouse.update();
			g_gui->update();

			// update camera rotation
			if (g_keyboard.isDown('A'))
			{
				g_targetRotation -= 0.1f;
			}
			else if (g_keyboard.isDown('D'))
			{
				g_targetRotation += 0.1f;
			}
			else if (g_mouse.isDown(MouseButtonRight))
			{
				if (!mouseRelative)
				{
					mouseRelative = true;

					g_mouse.getScreenPosition(mouseLastX, mouseLastY);
				}
				else
				{
					g_targetRotation += (g_mouse.getScreenPositionX() - mouseLastX) * 0.01f;
					g_mouse.getScreenPosition(mouseLastX, mouseLastY);
				}
			}
			else
			{
				if (mouseRelative)
				{
					mouseRelative = false;
				}
			}

			// update camera zoom
			if (g_keyboard.isDown('W'))
			{
				if (g_keyboard.isDown(Keyboard::LeftControl))
					g_cameraTargetDistance = MinCameraDistance;
				else
					g_cameraTargetDistance -= 30;
			}
			else if (g_keyboard.isDown('S'))
			{
				if (g_keyboard.isDown(Keyboard::LeftControl))
					g_cameraTargetDistance = MaxCameraDistance;
				else
					g_cameraTargetDistance += 30;
			}
			else
			{
				g_cameraTargetDistance -= g_mouse.getMovementZ() * 0.25f;
			}

			// limit zoom
			if (g_cameraTargetDistance < MinCameraDistance)
			{
				g_cameraTargetDistance = MinCameraDistance;
			}
			else if (g_cameraTargetDistance > MaxCameraDistance)
			{
				g_cameraTargetDistance = MaxCameraDistance;
			}

			// begin new frame by clearing screen to white
			glClearColor(1,1,1,0);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

			// use opengl to update camera matrix
			glLoadIdentity();

			g_cameraDistance += (g_cameraTargetDistance - g_cameraDistance) * g_system.getFrameTime();
			g_rotation += (g_targetRotation - g_rotation) * g_system.getFrameTime();

			g_cameraPosition = Vector3f(g_cameraDistance * sinf(g_rotation), g_cameraDistance * cosf(g_rotation),150);
			gluLookAt(g_cameraPosition[0], g_cameraPosition[1], g_cameraPosition[2], 0,0,0, 0,0,1);

			g_modelMatrix.identity();

			// copy camera matrices
			glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, &g_viewMatrix[0]);
			glGetFloatv(GL_TRANSPOSE_PROJECTION_MATRIX, &g_projectionMatrix[0]);

			// render scene (render.cpp)
			if (!render())
			{
				// here is where we end up if the .fx-file did not load...
			}

			// render GUI
			g_gui->render();

			// swap buffers
			g_window.swapBuffers();
		}
		
		// remove GUI
		delete g_gui;

		// shutdown and release rendering resources (render.cpp)
		shutdownRender();
	}

	// shutdown subsystems
	g_effectManager.destroy();
	g_textureManager.destroy();
	g_mouse.destroy();
	g_keyboard.destroy();
	g_window.destroy();
	g_console.destroy();
	g_system.destroy();
}
