
// includes
#include <cmath>
#include <fstream>
#include <iostream>
#include <map>

#include "bleifrei/gfx/opengl.hpp"
#include "bleifrei/gfx/Effect.hpp"
#include "bleifrei/gfx/EffectManager.hpp"
#include "bleifrei/img/TGAImage.hpp"
#include "bleifrei/io/Console.hpp"
#include "bleifrei/math/Matrix_op.hpp"
#include "bleifrei/scene/Mesh.hpp"
#include "bleifrei/scene/Scene.hpp"
#include "bleifrei/tex/Texture.hpp"
#include "bleifrei/tex/TextureManager.hpp"

#include "global.hpp"

using namespace bleifrei::gfx;
using namespace bleifrei::img;
using namespace bleifrei::io;
using namespace bleifrei::math;
using namespace bleifrei::platform;
using namespace bleifrei::scene;
using namespace bleifrei::tex;
using namespace bleifrei;
using namespace std;

static int progress = 0;

// global effect status flag
bool g_effectLoaded = false;

// global effect file name
string g_effectFilename = "carscene-1.fx";

// pointer to global effect (CgFX)
Effect *g_effect = 0;

// global scene pointer
Scene *g_scene = 0;

// global skybox texture pointer
TextureCube *g_environmentMap = 0;

TextureRect *g_backbufferTexture = 0;

// global techniques
Technique *g_techniqueSimple					= 0;
Technique *g_techniqueOpaque					= 0;
Technique *g_techniqueTransparent				= 0;

// global matrix parameters
Parameter *g_modelMatrixParam					= 0;
Parameter *g_viewProjectionMatrixParam			= 0;

// global texture parameters
Parameter *g_diffuseMapParam					= 0;
Parameter *g_environmentMapParam				= 0;

// global material parameters
Parameter *g_materialAmbientParam				= 0;
Parameter *g_materialDiffuseParam				= 0;
Parameter *g_materialSpecularParam				= 0;
Parameter *g_materialShininessParam				= 0;

// global light parameters
Parameter *g_light0AmbientParam					= 0;
Parameter *g_light0DiffuseParam					= 0;
Parameter *g_light0SpecularParam				= 0;
Parameter *g_light0PositionParam				= 0;

// global general effect parameters
Parameter *g_timeParam							= 0;
Parameter *g_cameraPositionParam				= 0;

// shading types
enum Shading
{
	ShadingSimple,
	ShadingOpaque,
	ShadingTransparent,
	ShadingUndefined
};

vector<Shading> g_shadingLookup;

// converts vector3f and alpha value to vector4f
inline Vector4f
addAlpha(const Vector3f &color, float alpha)
{
	return Vector4f(color[0], color[1], color[2], alpha);
}

void
assignShaders(Mesh *mesh)
{
	setLoadingText("Assigning shaders...");
	progressCallback(progress);

	multimap<Shading, unsigned int> shadingMap;
	for (unsigned int i = 0; i < mesh->getMaterialIndices().size(); ++i)
	{
		Material &material = g_scene->getMaterials()[mesh->getMaterialIndices()[i]];

		Shading shading;
		if (material.getDiffuseMap1() == g_textureManager.load("SHADOW.TGA"))
		{
			shading = ShadingSimple;
		}
		else
		{
			if (material.getOpacityMap() == material.getDiffuseMap1())
			{
				shading = ShadingTransparent;
			}
			else
			{
				shading = ShadingOpaque;
			}
		}
		
		shadingMap.insert(pair<Shading, unsigned int>(shading, i));
	}

	setLoadingText("Sorting triangles...");
	progressCallback(++progress);

	vector<Vector3ui> triangles;
	vector<unsigned int> materialIndices;

	multimap<Shading, unsigned int>::iterator it;
	for (it = shadingMap.begin(); it != shadingMap.end(); ++it)
	{
		triangles.push_back(mesh->getTriangles()[it->second]);
		materialIndices.push_back(mesh->getMaterialIndices()[it->second]);
		g_shadingLookup.push_back(it->first);
	}
	mesh->getTriangles() = triangles;
	mesh->getMaterialIndices() = materialIndices;
}

void
loadScene(void)
{
	Mesh *mergedMesh = new Mesh();
	mergedMesh->setName("EDA 425");

	g_scene = new Scene();
	g_scene->insert(mergedMesh);

	Scene *tempScene = new Scene();
	string sceneName = "media/car.scene";

	setLoadingText("Loading scene: " + sceneName);
	progressCallback(progress);
	tempScene->read(ifstream(sceneName.c_str(), ios::binary));

	// Copy materials: 1 step
	// Merge meshes: x steps
	// Assign shaders: 1 step
	// Sort triangles: 1 step
	// Load effect: 1 step
	// Load environment map: 1 step
	setLoadingProgressMax(1 + (int)tempScene->getMeshes().size() + 1 + 1 + 1 + 1);

	setLoadingText("Copying materials...");
	progressCallback(++progress);
	g_scene->getMaterials() = tempScene->getMaterials();

	vector<Mesh *> &meshes = tempScene->getMeshes();
	for (vector<Mesh *>::iterator it = meshes.begin(); it != meshes.end(); ++it)
	{
		setLoadingText("Merging mesh: " + (*it)->getName());
		progressCallback(++progress);
		mergedMesh->append(*it);
	}

	assignShaders(mergedMesh);

	safe_release(mergedMesh);
	safe_release(tempScene);
}

Technique *
getTechniquePointerByName(const string &name)
{
	Effect::technique_iterator tit = g_effect->getTechniqueByName(name);
	if (tit == g_effect->techniqueEnd())
	{
		g_console.printf(1, 0, 0, "ERROR: Failed to load technique \"%s\"!", name.c_str());
		g_effectLoaded = false;
		return 0;
	}
	return &(*tit);
}

Parameter *
getParameterPointerBySemantic(const string &semantic)
{
	Effect::parameter_iterator pit = g_effect->getParameterBySemantic(semantic);
	if (pit == g_effect->parameterEnd())
	{
		g_console.printf(1, 0, 0, "ERROR: Could not find parameter with semantic \"%s\"!", semantic.c_str());
		g_effectLoaded = false;
		return 0;
	}
	return &(*pit);
}

void
initEffect(void)
{
	g_effectLoaded = true;

	safe_release(g_effect);
	g_effect = g_effectManager.reload(g_effectFilename);
	if (!g_effect)
	{
		g_effectLoaded = false;
		return;
	}
	g_effect->addReference();

	// Get pointers to techniques
	g_techniqueSimple           = getTechniquePointerByName("simple");
	g_techniqueOpaque           = getTechniquePointerByName("opaque");
	g_techniqueTransparent      = getTechniquePointerByName("transparent");

	g_modelMatrixParam          = getParameterPointerBySemantic("MODEL_MATRIX");
	g_viewProjectionMatrixParam = getParameterPointerBySemantic("VIEW_PROJECTION_MATRIX");

	g_diffuseMapParam						= getParameterPointerBySemantic("DIFFUSE_MAP");
	g_environmentMapParam       = getParameterPointerBySemantic("ENVIRONMENT_MAP");

	g_materialAmbientParam      = getParameterPointerBySemantic("MATERIAL_AMBIENT");
	g_materialDiffuseParam      = getParameterPointerBySemantic("MATERIAL_DIFFUSE");
	g_materialSpecularParam     = getParameterPointerBySemantic("MATERIAL_SPECULAR");
	g_materialShininessParam    = getParameterPointerBySemantic("MATERIAL_SHININESS");

	g_light0AmbientParam        = getParameterPointerBySemantic("LIGHT0_AMBIENT");
	g_light0DiffuseParam        = getParameterPointerBySemantic("LIGHT0_DIFFUSE");
	g_light0SpecularParam       = getParameterPointerBySemantic("LIGHT0_SPECULAR");
	g_light0PositionParam       = getParameterPointerBySemantic("LIGHT0_POSITION");

	g_timeParam                 = getParameterPointerBySemantic("TIME");
	g_cameraPositionParam       = getParameterPointerBySemantic("CAMERA_POSITION");
}

void
setLightParameters(void)
{
	g_effect->setParameter(*g_light0AmbientParam, Vector4f(0.5f, 0.5f, 0.5f, 1.0f));
	g_effect->setParameter(*g_light0DiffuseParam, Vector4f(0.75f, 0.75f, 0.75f, 1.0f));
	g_effect->setParameter(*g_light0SpecularParam, Vector4f(0.25f, 0.25f, 0.25f, 1.0f));
	g_effect->setParameter(*g_light0PositionParam, Vector4f(0.0f, -200.0f, 200.0f, 1.0f));
}

void
setCameraParameters(void)
{
	Matrix4x4f viewProjectionMatrix = g_viewMatrix * g_projectionMatrix;

	g_effect->setParameter(*g_modelMatrixParam, g_modelMatrix);
	g_effect->setParameter(*g_viewProjectionMatrixParam, viewProjectionMatrix);
	g_effect->setParameter(*g_cameraPositionParam, g_cameraPosition); 
//	cgGLSetStateMatrixParameter(g_modelMatrixParam->cg_getParameter(), CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_IDENTITY);
//	cgGLSetStateMatrixParameter(g_viewProjectionMatrixParam->cg_getParameter(), CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
}

void
renderScene(void)
{
	// enable client states for vertex arrays
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	vector<Mesh *> &meshes = g_scene->getMeshes();

	// iterate over all meshes in scene
	vector<Mesh *>::iterator it;
	for (it = meshes.begin(); it != meshes.end(); ++it)
	{
		// set pointers to vertex data
		glVertexPointer(3, GL_FLOAT, 0, &(*it)->getVertices()[0]);
		glTexCoordPointer(2, GL_FLOAT, 0, &(*it)->getTexCoords()[0]);
		glNormalPointer(GL_FLOAT, 0, &(*it)->getNormals()[0]);

		// get first material index
		unsigned int materialIndex = (*it)->getMaterialIndices()[0];

		// iterate over all triangles
		unsigned int triangleCount = static_cast<unsigned int>((*it)->getTriangles().size());
		for (unsigned int i = 0; i < triangleCount;)
		{
			// get current material
			Material &material = g_scene->getMaterials()[materialIndex];
	
			// convert transparency value to alpha
			float alpha = 1 - material.getTransparency();

			// set material parameters
			g_effect->setParameter(*g_materialShininessParam, material.getShininess());
			g_effect->setParameter(*g_materialAmbientParam, addAlpha(material.getAmbientColor(), alpha));
			g_effect->setParameter(*g_materialDiffuseParam, addAlpha(material.getDiffuseColor(), alpha));
			g_effect->setParameter(*g_materialSpecularParam, addAlpha(material.getSpecularColor(), alpha));

			// get current diffusemap
			Texture *diffuseMap = material.getDiffuseMap1();

			// set technique and parameters (using lookup)
			switch (g_shadingLookup[i])
			{
			case ShadingSimple :
				g_effect->setTechnique(*g_techniqueSimple);
				g_effect->setParameter(*g_diffuseMapParam, material.getDiffuseMap1());
				break;
			case ShadingOpaque :
				g_effect->setTechnique(*g_techniqueOpaque);
				g_effect->setParameter(*g_diffuseMapParam, material.getDiffuseMap1());
				g_effect->setParameter(*g_environmentMapParam, g_environmentMap);
				break;
			case ShadingTransparent:
				g_effect->setTechnique(*g_techniqueTransparent);
				g_effect->setParameter(*g_diffuseMapParam, material.getDiffuseMap1());
				g_effect->setParameter(*g_environmentMapParam, g_environmentMap);
				break;
			}

			// find triangles in sequence with current material
			unsigned int first = i;
			for (; i < triangleCount; ++i)
			{
				unsigned int newIndex = (*it)->getMaterialIndices()[i];
				if (newIndex != materialIndex)
				{
					materialIndex = newIndex;
					break;
				}
			}

			unsigned int passCount = 0;
			
			// begin rendering
			g_effect->begin(passCount);
			{
				for (unsigned int pass = 0; pass < passCount; ++pass)
				{
					// begin pass
					g_effect->beginPass(pass);

					// render triangles
					glDrawElements(GL_TRIANGLES, 3 * (i - first), GL_UNSIGNED_INT, &(*it)->getTriangles()[first][0]);

					g_effect->endPass();
				}
			}
			g_effect->end();
		}
	}

	// disable client states
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}

// render scene
bool
render(void)
{
	if (!g_effectLoaded) return false;

	setLightParameters();
	setCameraParameters();

	g_effect->setParameter(*g_timeParam, g_time);

	renderScene();

	// Restore textures after evil shader code...
	nullTexture->forceBind();

	return true;
}

void
loadEnvironmentMap(void)
{
	setLoadingText("Loading environment cube map...");
	progressCallback(++progress);

	g_environmentMap = new TextureCube();

	try
	{
		TGAImage neg_x(ifstream("media/sky203.tga", ios::binary));
		TGAImage pos_x(ifstream("media/sky201.tga", ios::binary));
		TGAImage neg_y(ifstream("media/sky202.tga", ios::binary));
		TGAImage pos_y(ifstream("media/sky204.tga", ios::binary));
		TGAImage neg_z(ifstream("media/sky205.tga", ios::binary));
		TGAImage pos_z(ifstream("media/sky200.tga", ios::binary));

		g_environmentMap->load(neg_x, pos_x, neg_y, pos_y, neg_z, pos_z);
	}
	catch (...)
	{
		g_console.print(1, 0, 0, "ERROR: Could not load environment map!");
	}
}

void
initRender(void)
{
	progress = 0;
	loadScene();

	g_backbufferTexture = new TextureRect();

	setLoadingText("Loading effect file: " + g_effectFilename);
	progressCallback(++progress);

	initEffect();
	loadEnvironmentMap();
	setLoadingText("Done!");
	progressCallback(++progress);
}

void
shutdownRender(void)
{
	safe_release(g_environmentMap);
	safe_release(g_backbufferTexture);
	safe_release(g_scene);
	safe_release(g_effect);
}
